@@ -1,6 +1,6 @@
 ##############################################################################
 #
-# Copyright (c) 2006 Lovely Systems and Contributors.
+# Copyright (c) 2006, 2007 Lovely Systems and Contributors.
 # All Rights Reserved.
 #
 # This software is subject to the provisions of the Zope Public License,
@@ -13,7 +13,7 @@
 ##############################################################################
 """Task Service Implementation
 
-$Id: service.py 81533 2007-11-05 16:57:26Z fdrake $
+$Id: service.py 97411 2009-03-02 09:02:04Z zagy $
 """
 __docformat__ = 'restructuredtext'
 
@@ -24,25 +24,20 @@ import threading
 import time
 import zc.queue
 import zope.interface
-import zope.publisher.base
-import zope.publisher.publish
+import zope.location
 from BTrees.IOBTree import IOBTree
+from lovely.remotetask import interfaces, job, task, processor
 from zope import component
-from zope.app import zapi
 from zope.app.appsetup.product import getProductConfiguration
 from zope.app.container import contained
-from zope.app.component.interfaces import ISite
 from zope.app.publication.zopepublication import ZopePublication
-from zope.security.proxy import removeSecurityProxy
-from zope.traversing.api import traverse
 from zope.component.interfaces import ComponentLookupError
-from lovely.remotetask import interfaces, job, task
+from zope.traversing.api import getParents
 
 log = logging.getLogger('lovely.remotetask')
 
 storage = threading.local()
 
-
 class TaskService(contained.Contained, persistent.Persistent):
     """A persistent task service.
 
@@ -51,6 +46,8 @@ class TaskService(contained.Contained, persistent.Persistent):
     zope.interface.implements(interfaces.ITaskService)
 
     taskInterface = interfaces.ITask
+    processorFactory = processor.SimpleProcessor
+    processorArguments = {'waitTime': 1.0}
 
     _scheduledJobs  = None
     _scheduledQueue = None
@@ -59,15 +56,15 @@ class TaskService(contained.Contained, persistent.Persistent):
         super(TaskService, self).__init__()
         self._counter = 1
         self.jobs = IOBTree()
-        self._queue = zc.queue.PersistentQueue()
+        self._queue = zc.queue.Queue()
         self._scheduledJobs = IOBTree()
-        self._scheduledQueue = zc.queue.PersistentQueue()
+        self._scheduledQueue = zc.queue.Queue()
 
     def getAvailableTasks(self):
         """See interfaces.ITaskService"""
         return dict(component.getUtilitiesFor(self.taskInterface))
 
-    def add(self, task, input=None):
+    def add(self, task, input=None, startLater=False):
         """See interfaces.ITaskService"""
         if task not in self.getAvailableTasks():
             raise ValueError('Task does not exist')
@@ -75,8 +72,11 @@ class TaskService(contained.Contained, persistent.Persistent):
         self._counter += 1
         newjob = job.Job(jobid, task, input)
         self.jobs[jobid] = newjob
-        self._queue.put(newjob)
-        newjob.status = interfaces.QUEUED
+        if startLater:
+            newjob.status = interfaces.STARTLATER
+        else:
+            self._queue.put(newjob)
+            newjob.status = interfaces.QUEUED
         return jobid
 
     def addCronJob(self, task, input=None,
@@ -99,17 +99,25 @@ class TaskService(contained.Contained, persistent.Persistent):
         self._scheduledQueue.put(newjob)
         return jobid
 
+    def startJob(self, jobid):
+        job = self.jobs[jobid]
+        if job.status == interfaces.STARTLATER:
+            self._queue.put(job)
+            job.status = interfaces.QUEUED
+            return True
+        return False
+
     def reschedule(self, jobid):
         self._scheduledQueue.put(self.jobs[jobid])
 
-    def clean(self, stati=[interfaces.CANCELLED, interfaces.ERROR,
-                           interfaces.COMPLETED]):
+    def clean(self, status=[interfaces.CANCELLED, interfaces.ERROR,
+                            interfaces.COMPLETED]):
         """See interfaces.ITaskService"""
         allowed = [interfaces.CANCELLED, interfaces.ERROR,
                    interfaces.COMPLETED]
         for key in list(self.jobs.keys()):
             job = self.jobs[key]
-            if job.status in stati:
+            if job.status in status:
                 if job.status not in allowed:
                     raise ValueError('Not allowed status for removing. %s' % \
                         job.status)
@@ -126,6 +134,7 @@ class TaskService(contained.Contained, persistent.Persistent):
             job = self.jobs[jobid]
             if (   job.status == interfaces.CRONJOB
                 or job.status == interfaces.DELAYED
+                or job.status == interfaces.STARTLATER
                ):
                 job.status = interfaces.CANCELLED
 
@@ -149,15 +158,15 @@ class TaskService(contained.Contained, persistent.Persistent):
             self._scheduledJobs = IOBTree()
         if self._scheduledQueue == None:
             self._scheduledQueue = zc.queue.PersistentQueue()
-        path = [parent.__name__ for parent in zapi.getParents(self)
-                if parent.__name__]
-        path.reverse()
-        path.append(self.__name__)
-        path.append('processNext')
-
-        thread = threading.Thread(
-            target=processor, args=(self._p_jar.db(), path),
-            name=self._threadName())
+        # Create the path to the service within the DB.
+        servicePath = [parent.__name__ for parent in getParents(self)
+                       if parent.__name__]
+        servicePath.reverse()
+        servicePath.append(self.__name__)
+        # Start the thread running the processor inside.
+        processor = self.processorFactory(
+            self._p_jar.db(), servicePath, **self.processorArguments)
+        thread = threading.Thread(target=processor, name=self._threadName())
         thread.setDaemon(True)
         thread.running = True
         thread.start()
@@ -180,21 +189,52 @@ class TaskService(contained.Contained, persistent.Persistent):
                 if thread.getName() == name:
                     if thread.running:
                         return True
+                    break
         return False
 
     def _threadName(self):
         """Return name of the processing thread."""
         # This name isn't unique based on the path to self, but this doesn't
         # change the name that's been used in past versions.
-        path = [parent.__name__ for parent in zapi.getParents(self)
+        path = [parent.__name__ for parent in getParents(self)
                 if parent.__name__]
         path.append('remotetasks')
         path.reverse()
         path.append(self.__name__)
         return '.'.join(path)
 
-    def processNext(self, now=None):
+    def hasJobsWaiting(self, now=None):
+        # If there is are any simple jobs in the queue, we have work to do.
+        if self._queue:
+            return True
+        # First, move new cron jobs from the scheduled queue into the cronjob
+        # list.
+        if now is None:
+            now = int(time.time())
+        while len(self._scheduledQueue) > 0:
+            job = self._scheduledQueue.pull()
+            if job.status is not interfaces.CANCELLED:
+                self._insertCronJob(job, now)
+        # Now get all jobs that should be done now or earlier; if there are
+        # any that do not have errors or are cancelled, then we have jobs to
+        # do.
+        for key in self._scheduledJobs.keys(max=now):
+            jobs = [job for job in self._scheduledJobs[key]
+                    if job.status not in (interfaces.CANCELLED,
+                                          interfaces.ERROR)]
+            if jobs:
+                return True
+        return False
+
+    def claimNextJob(self, now=None):
         job = self._pullJob(now)
+        return job and job.id or None
+
+    def processNext(self, now=None, jobid=None):
+        if jobid is None:
+            job = self._pullJob(now)
+        else:
+            job = self.jobs[jobid]
         if job is None:
             return False
         try:
@@ -220,8 +260,8 @@ class TaskService(contained.Contained, persistent.Persistent):
                 job.status = interfaces.ERROR
         except Exception, error:
             if storage.runCount <= 3:
-                log.error(
-                    'catched a generic exception, preventing thread from crashing')
+                log.error('Caught a generic exception, preventing thread '
+                          'from crashing')
                 log.exception(error)
                 raise
             else:
@@ -242,7 +282,7 @@ class TaskService(contained.Contained, persistent.Persistent):
         # list
         if now is None:
             now = int(time.time())
-        while len(self._scheduledQueue)>0:
+        while len(self._scheduledQueue) > 0:
             job = self._scheduledQueue.pull()
             if job.status is not interfaces.CANCELLED:
                 self._insertCronJob(job, now)
@@ -290,33 +330,6 @@ class TaskService(contained.Contained, persistent.Persistent):
         self._scheduledJobs[nextCallTime] = jobs + (job,)
 
 
-class ProcessorPublication(ZopePublication):
-    """A custom publication to process the next job."""
-
-    def traverseName(self, request, ob, name):
-        return traverse(removeSecurityProxy(ob), name, None)
-
-
-def processor(db, path):
-    """Job Processor
-
-    Process the jobs that are waiting in the queue. This processor is meant to
-    be run in a separate process; however, it simply goes back to the task
-    service to actually do the processing.
-    """
-    path.reverse()
-    while threading.currentThread().running:
-        request = zope.publisher.base.BaseRequest(None, {})
-        request.setPublication(ProcessorPublication(db))
-        request.setTraversalStack(path)
-        try:
-            zope.publisher.publish.publish(request, False)
-            if not request.response._result:
-                time.sleep(1)
-        except:
-            # This thread should never crash, thus a blank except
-            pass
-
 def getAutostartServiceNames():
     """get a list of services to start"""
 
@@ -350,7 +363,7 @@ def bootStrapSubscriber(event):
             sites = []
             sites.append(root_folder)
             for folder in root_folder.values():
-                if ISite.providedBy(folder):
+                if zope.location.interfaces.ISite.providedBy(folder):
                     sites.append(folder)
         else:
             sites = [root_folder.get(siteName)]
